import re
import sre_constants
import yaml
from prospector.message import Location, Message
from prospector.profiles import AUTO_LOADED_PROFILES
from prospector.tools import ToolBase
from prospector.tools import pyflakes


CONFIG_SETTING_SHOULD_BE_LIST = 'should-be-list'
CONFIG_UNKNOWN_SETTING = 'unknown-setting'
CONFIG_SETTING_MUST_BE_INTEGER = 'should-be-int'
CONFIG_SETTING_MUST_BE_BOOL = 'should-be-bool'
CONFIG_INVALID_VALUE = 'invalid-value'
CONFIG_INVALID_REGEXP = 'invalid-regexp'
CONFIG_DEPRECATED_SETTING = 'deprecated'
CONFIG_DEPRECATED_CODE = 'deprecated-tool-code'


class ProfileValidationTool(ToolBase):

    LIST_SETTINGS = (
        'inherits', 'uses', 'ignore', 'ignore-paths', 'ignore-patterns'
    )
    ALL_SETTINGS = LIST_SETTINGS + (
        'strictness', 'autodetect', 'max-line-length',
        'output-format', 'doc-warnings', 'test-warnings',
        'requirements',  # bit of a grim hack; prospector does not use this but Landscape does
    )

    def __init__(self):
        self.to_check = set(AUTO_LOADED_PROFILES)
        self.ignore_codes = ()

    def configure(self, prospector_config, found_files):
        for profile in prospector_config.config.profiles:
            self.to_check.add(profile)

        self.ignore_codes = prospector_config.get_disabled_messages('profile-validator')

    def tool_names(self):
        # TODO: this is currently a circular import, which is why it is not at the top of
        # the module. However, there's no obvious way to get around this right now...
        # pylint: disable=cyclic-import
        from prospector.tools import TOOLS
        return TOOLS.keys()

    def validate(self, filepath):
        messages = []

        with open(filepath) as profile_file:
            raw_contents = profile_file.read()
            parsed = yaml.safe_load(raw_contents)
            raw_contents = raw_contents.split('\n')

        def add_message(code, message, setting):
            if code in self.ignore_codes:
                return
            line = -1
            for number, fileline in enumerate(raw_contents):
                if setting in fileline:
                    line = number + 1
                    break
            location = Location(filepath, None, None, line, 0, False)
            message = Message('profile-validator', code, location, message)
            messages.append(message)

        for setting in ('doc-warnings', 'test-warnings', 'autodetect'):
            if not isinstance(parsed.get(setting, False), bool):
                add_message(CONFIG_SETTING_MUST_BE_BOOL, '"%s" should be true or false' % setting, setting)

        if not isinstance(parsed.get('max-line-length', 0), int):
            add_message(CONFIG_SETTING_MUST_BE_INTEGER, '"max-line-length" should be an integer', 'max-line-length')

        if 'strictness' in parsed:
            possible = ('veryhigh', 'high', 'medium', 'low', 'verylow', 'none')
            if parsed['strictness'] not in possible:
                add_message(CONFIG_INVALID_VALUE, '"strictness" must be one of %s' % ', '.join(possible), 'strictness')

        if 'uses' in parsed:
            possible = ('django', 'celery')
            for uses in parsed['uses']:
                if uses not in possible:
                    add_message(CONFIG_INVALID_VALUE,
                                '"%s" is not valid for "uses", must be one of %s' % (uses, ', '.join(possible)),
                                uses)

        if 'ignore' in parsed:
            add_message(CONFIG_DEPRECATED_SETTING,
                        '"ignore" is deprecated, please update to use "ignore-patterns" instead',
                        'ignore')

        for pattern in parsed.get('ignore-patterns', []):
            try:
                re.compile(pattern)
            except sre_constants.error:
                add_message(CONFIG_INVALID_REGEXP, 'Invalid regular expression', pattern)

        for key in ProfileValidationTool.LIST_SETTINGS:
            if key not in parsed:
                continue
            if not isinstance(parsed[key], (tuple, list)):
                add_message(CONFIG_SETTING_SHOULD_BE_LIST, '"%s" should be a list' % key, key)

        for key in parsed.keys():
            if key not in ProfileValidationTool.ALL_SETTINGS and key not in self.tool_names():
                add_message(CONFIG_UNKNOWN_SETTING, '"%s" is not a valid prospector setting' % key, key)

        if 'pyflakes' in parsed:
            for code in parsed['pyflakes'].get('enable', []) + parsed['pyflakes'].get('disable', []):
                if code in pyflakes.LEGACY_CODE_MAP:
                    add_message(CONFIG_DEPRECATED_CODE,
                                'Pyflakes code %s was renamed to %s' % (code, pyflakes.LEGACY_CODE_MAP[code]),
                                'pyflakes')

        return messages

    def run(self, found_files):
        messages = []
        for filepath in found_files.iter_file_paths(abspath=False, include_ignored=True):
            if filepath in self.to_check:
                messages += self.validate(filepath)
        return messages



from __future__ import absolute_import

from pyflakes.api import checkPath
from pyflakes.reporter import Reporter

from prospector.message import Location, Message
from prospector.tools.base import ToolBase


__all__ = (
    'PyFlakesTool',
)


# Prospector uses the same pyflakes codes as flake8 defines,
# see http://flake8.readthedocs.org/en/latest/warnings.html
# and https://bitbucket.org/tarek/flake8/src/a209fb69350c572c9b2d7b4b09c7657be153be5e/flake8/_pyflakes.py
_MESSAGE_CODES = {
    'UnusedImport': 'F401',
    'ImportShadowedByLoopVar': 'F402',
    'ImportStarUsed': 'F403',
    'LateFutureImport': 'F404',
    'Redefined': 'F810',
    'RedefinedWhileUnused': 'F811',
    'RedefinedInListComp': 'F812',
    'UndefinedName': 'F821',
    'UndefinedExport': 'F822',
    'UndefinedLocal': 'F823',
    'DuplicateArgument': 'F831',
    'UnusedVariable': 'F841',
}

# The old prospector codes were deprecated in favour of using the codes
# defined by flake8. This maps the old codes to the new codes.
LEGACY_CODE_MAP = {
    'FL0001': 'F401',
    'FL0002': 'F811',
    'FL0003': 'F812',
    'FL0004': 'F402',
    'FL0005': 'F403',
    'FL0006': 'F821',
    'FL0008': 'F822',
    'FL0009': 'F823',
    'FL0010': 'F831',
    'FL0011': 'F810',
    'FL0012': 'F404',
    'FL0013': 'F841',
}


class ProspectorReporter(Reporter):
    def __init__(self, ignore=None):
        super(ProspectorReporter, self).__init__(None, None)
        self._messages = []
        self.ignore = ignore or ()

    # pylint: disable=too-many-arguments
    def record_message(
            self,
            filename=None,
            line=None,
            character=None,
            code=None,
            message=None):

        code = code or 'F999'
        if code in self.ignore:
            return

        location = Location(
            path=filename,
            module=None,
            function=None,
            line=line,
            character=character,
        )
        message = Message(
            source='pyflakes',
            code=code,
            location=location,
            message=message,
        )
        self._messages.append(message)

    def unexpectedError(self, filename, msg):  # noqa
        self.record_message(
            filename=filename,
            code='F999',
            message=msg,
        )

    # pylint: disable=too-many-arguments
    def syntaxError(self, filename, msg, lineno, offset, text):  # noqa
        self.record_message(
            filename=filename,
            line=lineno,
            character=offset,
            code='F999',
            message=msg,
        )

    def flake(self, message):
        code = _MESSAGE_CODES.get(message.__class__.__name__, 'F999')

        self.record_message(
            filename=message.filename,
            line=message.lineno,
            character=(message.col + 1),
            code=code,
            message=message.message % message.message_args,
        )

    def get_messages(self):
        return self._messages


class PyFlakesTool(ToolBase):
    def __init__(self, *args, **kwargs):
        super(PyFlakesTool, self).__init__(*args, **kwargs)
        self.ignore_codes = ()

    def configure(self, prospector_config, _):
        ignores = prospector_config.get_disabled_messages('pyflakes')
        # convert old style to new
        self.ignore_codes = (LEGACY_CODE_MAP.get(code, code) for code in ignores)
        return None

    def run(self, found_files):
        reporter = ProspectorReporter(ignore=self.ignore_codes)
        for filepath in found_files.iter_module_paths():
            checkPath(filepath, reporter)

        return reporter.get_messages()



from __future__ import absolute_import

import ast
from mccabe import PathGraphingAstVisitor

from prospector.message import Location, Message
from prospector.tools.base import ToolBase


__all__ = (
    'McCabeTool',
)


class McCabeTool(ToolBase):
    def __init__(self, *args, **kwargs):
        super(McCabeTool, self).__init__(*args, **kwargs)
        self.ignore_codes = ()
        self.max_complexity = 10

    def configure(self, prospector_config, _):
        self.ignore_codes = prospector_config.get_disabled_messages('mccabe')

        options = prospector_config.tool_options('mccabe')
        if 'max-complexity' in options:
            self.max_complexity = options['max-complexity']

        return None

    def run(self, found_files):
        messages = []

        for code_file in found_files.iter_module_paths():
            try:
                tree = ast.parse(
                    open(code_file, 'r').read(),
                    filename=code_file,
                )
            except (SyntaxError, TypeError):
                location = Location(
                    path=code_file,
                    module=None,
                    function=None,
                    line=1,
                    character=0,
                )
                message = Message(
                    source='mccabe',
                    code='MC0000',
                    location=location,
                    message='Could not parse file',
                )
                messages.append(message)
                continue

            visitor = PathGraphingAstVisitor()
            visitor.preorder(tree, visitor)

            for graph in visitor.graphs.values():
                complexity = graph.complexity()
                if complexity > self.max_complexity:
                    location = Location(
                        path=code_file,
                        module=None,
                        function=graph.entity,
                        line=graph.lineno,
                        character=0,
                        absolute_path=True
                    )
                    message = Message(
                        source='mccabe',
                        code='MC0001',
                        location=location,
                        message='%s is too complex (%s)' % (
                            graph.entity,
                            complexity,
                        ),
                    )
                    messages.append(message)

        return self.filter_messages(messages)

    def filter_messages(self, messages):
        return [
            message
            for message in messages
            if message.code not in self.ignore_codes
        ]



import codecs
import warnings
import os
import re
from requirements_detector import find_requirements
from requirements_detector.detect import RequirementsNotFound


POSSIBLE_LIBRARIES = ('django', 'celery')


# see http://docs.python.org/2/reference/lexical_analysis.html#identifiers
_FROM_IMPORT_REGEX = re.compile(r'^\s*from ([\._a-zA-Z0-9]+) import .*$')
_IMPORT_REGEX = re.compile(r'^\s*import ([\._a-zA-Z0-9]+)$')
_IMPORT_MULTIPLE_REGEX = re.compile(r'^\s*import ([\._a-zA-Z0-9]+(, ){1})+')
_CODING_REGEX = re.compile(r'coding[:=]\s*([-\w.]+)')


def detect_by_bom(path):
    """
    Determine and return file encoding using BOM character.

    This is necessary for files such as setuptools' tests/script-with-bom.py

    :returns bom-detected encoding only if discovered
    """
    with open(path, 'rb') as fin:
        raw = fin.read(4)
    for enc, boms in (
            ('utf-8-sig', (codecs.BOM_UTF8,)),
            ('utf-16', (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE)),
            ('utf-32', (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE))):
        if any(raw.startswith(bom) for bom in boms):
            return enc


def determine_pyfile_encoding(path, default='utf8'):
    """
    Determine and return file encoding of a python source file.

    https://www.python.org/dev/peps/pep-0263/

    :returns: file encoding if determined, otherwise ``default``.
    """
    encoding = detect_by_bom(path)
    if encoding:
        # trust the byte-order mark
        return encoding

    with open(path, 'rb') as fip:
        # look for string of form '# coding: <encoding>'
        # in the first two lines of python source file.
        for line in (_line for _line in
                     (fip.readline(), fip.readline())
                     if _line.decode(default).startswith('#')):
            match = next(re.finditer(_CODING_REGEX, line.decode('ascii')), None)
            if match:
                return match.group(1)
        return default


def find_from_imports(file_contents):
    names = set()
    for line in file_contents.split('\n'):
        match = _IMPORT_MULTIPLE_REGEX.match(line)
        if match:
            import_names = []
            first = match.group(1)
            import_names.append(first[:-2])
            for name in line.split(first)[1].split(','):
                import_names.append(name.strip())
        else:
            match = _IMPORT_REGEX.match(line) or _FROM_IMPORT_REGEX.match(line)
            if match is None:
                continue
            import_names = match.group(1).split('.')

        for import_name in import_names:
            if import_name in POSSIBLE_LIBRARIES:
                names.add(import_name)

    return names


def find_from_path(path):
    names = set()
    max_possible = len(POSSIBLE_LIBRARIES)

    for item in os.listdir(path):
        item_path = os.path.abspath(os.path.join(path, item))
        if os.path.isdir(item_path):
            names |= find_from_path(item_path)
        elif not os.path.islink(item_path) and item_path.endswith('.py'):
            try:
                encoding = determine_pyfile_encoding(item_path, default='utf8')
                with codecs.open(item_path, encoding=encoding) as fip:
                    names |= find_from_imports(fip.read())
            except UnicodeDecodeError as err:
                # this warning is issued: (1) in determine_pyfile_encoding for
                # badly authored files (contains non-utf8 in a comment line), or
                # in find_from_imports() for either (2) a coding is specified,
                # but wrong and (3) no coding is specified, and the default
                # 'utf8' fails to decode.
                warnings.warn('{0}: {1}'.format(item_path, err), ImportWarning)
            except LookupError as err:
                # an invalid 'coding' statement specified in target source file
                warnings.warn('{0}: {1}'.format(item_path, err), ImportWarning)

        if len(names) == max_possible:
            # don't continue on recursing, there's no point!
            break

    return names


def find_from_requirements(path):
    reqs = find_requirements(path)
    names = []
    for requirement in reqs:
        if (requirement.name is not None
                and requirement.name.lower() in POSSIBLE_LIBRARIES):
            names.append(requirement.name.lower())
    return names


def autodetect_libraries(path):

    if os.path.isfile(path):
        path = os.path.dirname(path)
        if path == '':
            path = '.'

    libraries = []

    try:
        libraries = find_from_requirements(path)

    # pylint: disable=pointless-except
    except RequirementsNotFound:
        pass

    if len(libraries) < len(POSSIBLE_LIBRARIES):
        libraries = find_from_path(path)

    return libraries





class ToolBase(object):
    # This is an 'abstract' base class, used to provide an indication of
    # how to create a new tool class. Therefore, the arguments will be unused,
    # so that is suppressed here.
    # pylint: disable=unused-argument

    def configure(self, prospector_config, found_files):
        """
        Tools have their own way of being configured from configuration files
        on the current path - for example, a .pep8rc file. Prospector will use
        its own configuration settings unless this method discovers some
        tool-specific configuration that should be used instead.

        :return: A string indicating how or where this tool was configured from.
                 For example, this can be a path to the .pylintrc file used, if
                 used. Returning None means that prospector defaults were used.
        """
        # We don't want Pylint to tell us that this method should be a function,
        # it's an "abstract" class after all.
        # pylint: disable=no-self-use
        return None

    def run(self, found_files):
        """
        Actually run the tool and collect the various messages emitted by the tool.
        It is expected that this will convert whatever output of the tool into the
        standard prospector Message and Location objects.
        """
        raise NotImplementedError



from __future__ import absolute_import
from pylint.reporters import BaseReporter
from pylint.utils import UnknownMessage
from prospector.message import Location, Message


class Collector(BaseReporter):

    name = 'collector'

    def __init__(self, message_store):
        BaseReporter.__init__(self, output=None)
        self._message_store = message_store
        self._messages = []

    def add_message(self, msg_id, location, msg):
        # (* magic is acceptable here)
        loc = Location(*location)
        # At this point pylint will give us the code but we want the
        # more user-friendly symbol
        try:
            msg_data = self._message_store.check_message_id(msg_id)
        except UnknownMessage:
            # this shouldn't happen, as all pylint errors should be
            # in the message store, but just in case we'll fall back
            # to using the code.
            msg_symbol = msg_id
        else:
            msg_symbol = msg_data.symbol

        message = Message('pylint', msg_symbol, loc, msg)
        self._messages.append(message)

    def _display(self, layout):
        pass

    def get_messages(self):
        return self._messages



from prospector.formatters import json, text, grouped, emacs, yaml, pylint


__all__ = (
    'FORMATTERS',
)


FORMATTERS = {
    'json': json.JsonFormatter,
    'text': text.TextFormatter,
    'grouped': grouped.GroupedFormatter,
    'emacs': emacs.EmacsFormatter,
    'yaml': yaml.YamlFormatter,
    'pylint': pylint.PylintFormatter,
}



from __future__ import absolute_import
import os.path
import sys

from datetime import datetime

from prospector import tools, blender, postfilter
from prospector.config import ProspectorConfig, configuration as cfg
from prospector.finder import find_python
from prospector.formatters import FORMATTERS
from prospector.message import Location, Message


__all__ = (
    'Prospector',
    'main',
)


class Prospector(object):
    def __init__(self, config):
        self.config = config
        self.summary = None
        self.messages = None

    def process_messages(self, found_files, messages):
        for message in messages:
            if self.config.absolute_paths:
                message.to_absolute_path(self.config.workdir)
            else:
                message.to_relative_path(self.config.workdir)
        if self.config.blending:
            messages = blender.blend(messages)

        filepaths = found_files.iter_module_paths(abspath=False)
        return postfilter.filter_messages(filepaths, messages)

    def execute(self):

        summary = {
            'started': datetime.now(),
        }
        summary.update(self.config.get_summary_information())

        found_files = find_python(self.config.ignores, self.config.paths,
                                  self.config.explicit_file_mode, self.config.workdir)

        # Run the tools
        messages = []
        for tool in self.config.get_tools(found_files):
            try:
                messages += tool.run(found_files)
            except Exception:  # pylint: disable=broad-except
                if self.config.die_on_tool_error:
                    raise
                else:
                    for name, cls in tools.TOOLS.items():
                        if cls == tool.__class__:
                            toolname = name
                            break
                    else:
                        toolname = 'Unknown'

                    loc = Location(self.config.workdir, None, None, None, None)
                    msg = 'Tool %s failed to run (exception was raised)' % (
                        toolname,
                    )
                    message = Message(
                        toolname,
                        'failure',
                        loc,
                        message=msg,
                    )
                    messages.append(message)

        messages = self.process_messages(found_files, messages)

        summary['message_count'] = len(messages)
        summary['completed'] = datetime.now()

        # Timedelta.total_seconds() is not available
        # on Python<=2.6 so we calculate it ourselves
        # See issue #60 and http://stackoverflow.com/a/3694895
        delta = (summary['completed'] - summary['started'])
        total_seconds = (delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 1e6) / 1e6
        summary['time_taken'] = '%0.2f' % total_seconds

        self.summary = summary
        self.messages = messages

    def get_summary(self):
        return self.summary

    def get_messages(self):
        return self.messages

    def print_messages(self, write_to=None):
        write_to = write_to or sys.stdout

        output_format = self.config.get_output_format()
        self.summary['formatter'] = output_format
        formatter = FORMATTERS[output_format](self.summary, self.messages, self.config.profile)

        print_messages = not self.config.summary_only and self.messages

        # Produce the output
        write_to.write(formatter.render(
            summary=not self.config.messages_only,
            messages=print_messages,
            profile=self.config.show_profile
        ).encode('utf-8'))
        write_to.write('\n')


def get_parser():
    """
    This is a helper method to return an argparse parser, to
    be used with the Sphinx argparse plugin for documentation.
    """
    manager = cfg.build_manager()
    source = cfg.build_command_line_source(prog='prospector', description=None)
    return source.build_parser(manager.settings, None)


def main():
    # Get our configuration
    config = ProspectorConfig()

    paths = config.paths
    if len(paths) > 1 and not all([os.path.isfile(path) for path in paths]):
        sys.stderr.write('\nIn multi-path mode, all inputs must be files, '
                         'not directories.\n\n')
        get_parser().print_usage()
        sys.exit(2)

    # Make it so
    prospector = Prospector(config)
    prospector.execute()
    prospector.print_messages()

    if config.exit_with_zero_on_success():
        # if we ran successfully, and the user wants us to, then we'll
        # exit cleanly
        return 0

    # otherwise, finding messages is grounds for exiting with an error
    # code, to make it easier for bash scripts and similar situations
    # to know if there any errors have been found.
    if len(prospector.get_messages()) > 0:
        return 1
    return 0


if __name__ == '__main__':
    sys.exit(main())



from __future__ import absolute_import

import tokenize
from pylint.checkers import BaseTokenChecker
from pylint.interfaces import ITokenChecker


class IndentChecker(BaseTokenChecker):

    __implements__ = (ITokenChecker,)

    name = 'indentation'
    msgs = {
        'W0313': ('File contains mixed indentation - some lines use tabs, some lines use spaces',
                  'indentation-mixture',
                  'Used when there are some mixed tabs and spaces in a module.'),
        'W0314': ('Line uses %s for indentation, but %s required',
                  'incorrect-indentation',
                  'Used when the indentation of a line does not match the style required by configuration.')
    }
    options = (
        ('indent-strict-spaces',
         {'default': False, 'type': "yn", 'metavar': '<boolean>',
          'help': 'Enforce using only spaces for indentation'}),

        ('indent-strict-tabs',
         {'default': False, 'type': "yn", 'metavar': '<boolean>',
          'help': 'Enforce using only tabs for indentation'})
    )

    def process_tokens(self, tokens):
        tab_count = space_count = 0
        line_num = 0

        for token in tokens:
            if token[0] == tokenize.NEWLINE:
                line_num += 1
                line = token[4]
                if line.startswith('\t'):
                    if self.config.indent_strict_spaces:
                        # we have tabs but are configured to only allow spaces
                        self.add_message('incorrect-indentation', line=line_num, args=('tabs', 'spaces'))
                    tab_count += 1

                if line.startswith(' '):
                    if self.config.indent_strict_tabs:
                        # we have tabs but are configured to only allow spaces
                        self.add_message('incorrect-indentation', line=line_num, args=('spaces', 'tabs'))
                    space_count += 1

        if tab_count > 0 and space_count > 0:
            # this file has mixed indentation!
            self.add_message('indentation-mixture', line=-1)



import json
from prospector.tools import TOOLS

import os
import yaml


class ProfileNotFound(Exception):
    def __init__(self, name, profile_path):
        super(ProfileNotFound, self).__init__()
        self.name = name
        self.profile_path = profile_path

    def __repr__(self):
        return "Could not find profile %s; searched in %s" % (self.name, ':'.join(self.filepath))


class ProspectorProfile(object):

    def __init__(self, name, profile_dict, inherit_order):
        self.name = name
        self.inherit_order = inherit_order

        self.ignore_paths = profile_dict.get('ignore-paths', [])
        # The 'ignore' directive is an old one which should be deprecated at some point
        self.ignore_patterns = profile_dict.get('ignore-patterns', []) + profile_dict.get('ignore', [])

        self.output_format = profile_dict.get('output-format')
        self.member_warnings = profile_dict.get('member-warnings')
        self.autodetect = profile_dict.get('autodetect')
        self.uses = [uses for uses in _ensure_list(profile_dict.get('uses', [])) if uses in ('django', 'celery')]
        self.max_line_length = profile_dict.get('max-line-length')

        for tool in TOOLS.keys():
            conf = {
                'disable': [],
                'enable': [],
                'run': None,
                'options': {}
            }
            conf.update(profile_dict.get(tool, {}))

            if self.max_line_length is not None and tool in ('pylint', 'pep8'):
                conf['options']['max-line-length'] = self.max_line_length

            setattr(self, tool, conf)

    def get_disabled_messages(self, tool_name):
        disable = getattr(self, tool_name)['disable']
        enable = getattr(self, tool_name)['enable']
        return list(set(disable) - set(enable))

    def is_tool_enabled(self, name):
        return getattr(self, name).get('run')

    def list_profiles(self):
        # this profile is itself included
        return self.inherit_order

    def as_dict(self):
        out = {
            'ignore-paths': self.ignore_paths,
            'ignore-patterns': self.ignore_patterns,
            'output-format': self.output_format,
            'autodetect': self.autodetect,
            'uses': self.uses,
            'max-line-length': self.max_line_length,
            'member-warnings': self.member_warnings,
        }
        for tool in TOOLS.keys():
            out[tool] = getattr(self, tool)
        return out

    def as_json(self):
        return json.dumps(self.as_dict())

    def as_yaml(self):
        return yaml.safe_dump(self.as_dict())

    @staticmethod
    def load(name_or_path, profile_path, allow_shorthand=True, forced_inherits=None):
        # First simply load all of the profiles and those that it explicitly inherits from
        data, inherits = _load_and_merge(name_or_path, profile_path, allow_shorthand,
                                         forced_inherits=forced_inherits or [])
        return ProspectorProfile(name_or_path, data, inherits)


def _is_valid_extension(filename):
    ext = os.path.splitext(filename)[1]
    return ext in ('.yml', '.yaml')


def _load_content(name_or_path, profile_path):
    if _is_valid_extension(name_or_path) and os.path.exists(name_or_path):
        # assume that this is a full path that we can load
        filename = name_or_path
    else:
        filename = None
        for path in profile_path:
            for ext in ('yml', 'yaml'):
                filepath = os.path.join(path, '%s.%s' % (name_or_path, ext))
                if os.path.exists(filepath):
                    filename = filepath
                    break

        if filename is None:
            raise ProfileNotFound(name_or_path, profile_path)

    with open(filename) as fct:
        return yaml.safe_load(fct) or {}


def _ensure_list(value):
    if isinstance(value, list):
        return value
    return [value]


def _simple_merge_dict(priority, base):
    out = dict(base.items())
    out.update(dict(priority.items()))
    return out


def _merge_tool_config(priority, base):
    out = dict(base.items())
    for key, value in priority.items():
        if key in ('run', 'full', 'none'):  # pep8 has extra 'full' and 'none' options
            out[key] = value
        elif key in ('options',):
            out[key] = _simple_merge_dict(value, base.get(key, {}))

    # anything enabled in the 'priority' dict is removed
    # from 'disabled' in the base dict and vice versa
    base_disabled = base.get('disable') or []
    base_enabled = base.get('enable') or []
    pri_disabled = priority.get('disable') or []
    pri_enabled = priority.get('enable') or []

    out['disable'] = list(set(pri_disabled) | (set(base_disabled) - set(pri_enabled)))
    out['enable'] = list(set(pri_enabled) | (set(base_enabled) - set(pri_disabled)))

    return out


def _merge_profile_dict(priority, base):
    # copy the base dict into our output
    out = dict(base.items())

    for key, value in priority.items():
        if key in ('strictness', 'doc-warnings', 'test-warnings', 'member-warnings',
                   'output-format', 'autodetect', 'max-line-length',):
            # some keys are simple values which are overwritten
            out[key] = value
        elif key in ('ignore', 'ignore-patterns', 'ignore-paths', 'uses'):
            # some keys should be appended
            out[key] = _ensure_list(value) + _ensure_list(base.get(key, []))
        elif key in TOOLS.keys():
            # this is tool config!
            out[key] = _merge_tool_config(value, base.get(key, {}))

    return out


def _determine_strictness(profile_dict, inherits):
    for profile in inherits:
        if profile.startswith('strictness_'):
            return None, False

    strictness = profile_dict.get('strictness')
    if strictness is None:
        return None, False
    return ('strictness_%s' % strictness), True


def _determine_pep8(profile_dict):
    pep8 = profile_dict.get('pep8', {})
    if pep8.get('full', False):
        return 'full_pep8', True
    elif pep8.get('none', False):
        return 'no_pep8', True
    return None, False


def _determine_doc_warnings(profile_dict):
    doc_warnings = profile_dict.get('doc-warnings')
    if doc_warnings is None:
        return None, False
    return ('doc_warnings' if doc_warnings else 'no_doc_warnings'), True


def _determine_test_warnings(profile_dict):
    test_warnings = profile_dict.get('test-warnings')
    if test_warnings is None:
        return None, False
    return (None if test_warnings else 'no_test_warnings'), True


def _determine_member_warnings(profile_dict):
    member_warnings = profile_dict.get('member-warnings')
    if member_warnings is None:
        return None, False
    return ('member_warnings' if member_warnings else 'no_member_warnings'), True


def _determine_implicit_inherits(profile_dict, already_inherits, shorthands_found):
    # Note: the ordering is very important here - the earlier items
    # in the list have precedence over the later items. The point of
    # the doc/test/pep8 profiles is usually to restore items which were
    # turned off in the strictness profile, so they must appear first.
    implicit = [
        ('pep8', _determine_pep8(profile_dict)),
        ('docs', _determine_doc_warnings(profile_dict)),
        ('tests', _determine_test_warnings(profile_dict)),
        ('strictness', _determine_strictness(profile_dict, already_inherits)),
        ('members', _determine_member_warnings(profile_dict)),
    ]
    inherits = []

    for shorthand_name, determined in implicit:
        if shorthand_name in shorthands_found:
            continue
        extra_inherits, shorthand_found = determined
        if not shorthand_found:
            continue
        shorthands_found.add(shorthand_name)
        if extra_inherits is not None:
            inherits.append(extra_inherits)

    return inherits, shorthands_found


def _append_profiles(name, profile_path, data, inherit_list, allow_shorthand=False):
    new_data, new_il, _ = _load_profile(name, profile_path, allow_shorthand=allow_shorthand)
    data.update(new_data)
    inherit_list += new_il
    return data, inherit_list


def _load_and_merge(name_or_path, profile_path, allow_shorthand=True, forced_inherits=None):
    # First simply load all of the profiles and those that it explicitly inherits from
    data, inherit_list, shorthands_found = _load_profile(name_or_path, profile_path,
                                                         allow_shorthand=allow_shorthand,
                                                         forced_inherits=forced_inherits or [])

    if allow_shorthand:
        if 'docs' not in shorthands_found:
            data, inherit_list = _append_profiles('no_doc_warnings', profile_path, data, inherit_list)

        if 'members' not in shorthands_found:
            data, inherit_list = _append_profiles('no_member_warnings', profile_path, data, inherit_list)

        if 'tests' not in shorthands_found:
            data, inherit_list = _append_profiles('no_test_warnings', profile_path, data, inherit_list)

        if 'strictness' not in shorthands_found:
            # if no strictness was specified, then we should manually insert the medium strictness
            for inherit in inherit_list:
                if inherit.startswith('strictness_'):
                    break
            else:
                data, inherit_list = _append_profiles('strictness_medium', profile_path, data, inherit_list)

    # Now we merge all of the values together, from 'right to left' (ie, from the
    # top of the inheritance tree to the bottom). This means that the lower down
    # values overwrite those from above, meaning that the initially provided profile
    # has precedence.
    merged = {}
    for name in inherit_list[::-1]:
        priority = data[name]
        merged = _merge_profile_dict(priority, merged)

    return merged, inherit_list


def _load_profile(name_or_path, profile_path, shorthands_found=None,
                  already_loaded=None, allow_shorthand=True, forced_inherits=None):
    # recursively get the contents of the basic profile and those it inherits from
    base_contents = _load_content(name_or_path, profile_path)

    inherit_order = [name_or_path]
    shorthands_found = shorthands_found or set()

    already_loaded = already_loaded or []
    already_loaded.append(name_or_path)

    inherits = _ensure_list(base_contents.get('inherits', []))
    if forced_inherits is not None:
        inherits += forced_inherits

    # There are some 'shorthand' options in profiles which implicitly mean that we
    # should inherit from some of prospector's built-in profiles
    if base_contents.get('allow-shorthand', True) and allow_shorthand:
        extra_inherits, extra_shorthands = _determine_implicit_inherits(base_contents, inherits, shorthands_found)
        inherits += extra_inherits
        shorthands_found |= extra_shorthands

    contents_dict = {name_or_path: base_contents}

    for inherit_profile in inherits:
        if inherit_profile in already_loaded:
            # we already have this loaded and in the list
            continue

        already_loaded.append(inherit_profile)
        new_cd, new_il, new_sh = _load_profile(inherit_profile, profile_path,
                                               shorthands_found, already_loaded, allow_shorthand)
        contents_dict.update(new_cd)
        inherit_order += new_il
        shorthands_found |= new_sh

    return contents_dict, inherit_order, shorthands_found
